@@ -218,7 +218,9 @@ class ObserverPipeline:
         tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
         actions_taken: dict[str, float | int | None],
         return_dict: bool = False,
-    ) -> tuple[dict[str, list[float | int]], bool, dict[str, dict[str, list[float | int]]] | None]:
+    ) -> tuple[
+        dict[str, list[float | int]], bool, dict[str, dict[str, list[float | int] | dict[str, float | int]]] | None
+    ]:
         """
         :param observation_inputs: Observation input metrics not calculated with statistic trackers.
         :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
@@ -226,8 +228,10 @@ class ObserverPipeline:
         names to floats or TensorStatistic models.
         :param actions_taken: Dictionary mapping agent IDs to actions taken by that agent.
         :param return_dict: Whether to return a dictionary of observations as well as the normal vector.
-        :return: Tuple of a dictionary mapping agent ID to that agent's completed observation vector, a boolean
-        indicating whether an observation clip occurred and a dictionary of observations.
+        :return: Tuple of:
+         - A dictionary mapping agent ID to that agent's completed observation vector,
+         - A boolean indicating whether an observation clip occurred,
+         - A dictionary mapping agent ID to a dictionary mapping observer name to that observer's observation vector.
         """
 
         global_obs, global_obs_dict = self.global_observers.observe(
@@ -239,7 +243,7 @@ class ObserverPipeline:
         )
 
         local_obs: dict[str, list[float | int]] = {}
-        obs_as_dict: dict[str, dict[str, list[float | int]]] = {}
+        obs_as_dict: dict[str, dict[str, list[float | int] | dict[str, float | int]]] = {}
 
         for agent_id, agent_observers in self.local_observers.items():
             local_obs[agent_id], local_obs_dict = agent_observers.observe(
