@@ -8,6 +8,9 @@ from abc import ABC, abstractmethod
 from copy import deepcopy
 from typing import Any, TypeAlias, final
 
+import numpy as np
+from loguru import logger
+
 from libinephany.observations import observation_utils
 from libinephany.observations.observation_utils import StatisticStorageTypes
 from libinephany.pydantic_models.configs.observer_config import ObserverConfig
@@ -43,15 +46,18 @@ class Observer(ABC):
         standardizer: Standardizer | None,
         observer_config: ObserverConfig,
         should_standardize: bool = True,
-        skip_statistics: list[str] | None = None,
+        include_statistics: list[str] | None = None,
+        include_hparams: list[str] | None = None,
         **kwargs,
     ) -> None:
         """
         :param standardizer: None or the standardizer to apply to the returned observations.
         :param global_config: ObserverConfig that can be used to inform various observation calculations.
         :param should_standardize: Whether standardization should be applied to returned values.
-        :param skip_statistics: If the observation uses the TensorStatistic model to return observations, names of the
-        fields in the model to not include in returned observations.
+        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
+        fields in the model to include in returned observations.
+        :param include_hparams: If the observation uses the HyperparameterStates model to return observations, names of the
+        hyperparameters to include in returned observations.
         :param kwargs: Miscellaneous keyword arguments.
         """
 
@@ -62,8 +68,19 @@ class Observer(ABC):
         self.observer_config = observer_config
         self.standardize = standardizer if standardizer is not None else observation_utils.null_standardizer
         self.should_standardize = should_standardize and self.can_standardize
+        self.invalid_observation_threshold = observer_config.invalid_observation_threshold
 
-        self.skip_statistics = TensorStatistics.filter_skip_statistics(skip_statistics=skip_statistics)
+        self.include_statistics: list[str] | None = None
+        self.include_hparams = include_hparams
+
+        if include_statistics is not None:
+            self.include_statistics = TensorStatistics.filter_include_statistics(include_statistics=include_statistics)
+
+        if self.requires_include_statistics and not self.include_statistics:
+            raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")
+
+        if self.requires_include_hparams and not self.include_hparams:
+            raise ValueError(f"{self.__class__.__name__} must be provided with include_hparams.")
 
     @final
     @property
@@ -102,7 +119,10 @@ class Observer(ABC):
         observation_format = self.observation_format
 
         if observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
-            return len([field for field in TensorStatistics.model_fields.keys() if field not in self.skip_statistics])
+            if self.include_statistics is None:
+                raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")
+
+            return len([field for field in TensorStatistics.model_fields.keys() if field in self.include_statistics])
 
         elif observation_format is StatisticStorageTypes.FLOAT:
             return 1
@@ -137,6 +157,22 @@ class Observer(ABC):
 
         return True
 
+    @property
+    def requires_include_statistics(self) -> bool:
+        """
+        :return: Whether the observation requires include_statistics to be provided.
+        """
+
+        return False
+
+    @property
+    def requires_include_hparams(self) -> bool:
+        """
+        :return: Whether the observation requires include_hparams to be provided.
+        """
+
+        return False
+
     @property
     @abstractmethod
     def standardizer_key_infix(self) -> str:
@@ -188,6 +224,48 @@ class Observer(ABC):
 
         self._validated_observation = True
 
+    @final
+    def _validate_observation_values(self, observations: list[float | int]) -> None:
+        """
+        :param observations: Observation vector to validate.
+        """
+        validate = True
+        if len(observations) == 0:
+            raise ValueError(f"Observer: {self.__class__.__name__} gathered observations with empty list!")
+
+        if any(not isinstance(observation, (float, int)) for observation in observations):
+            # Just in case
+            other_count = sum(1 for observation in observations if not isinstance(observation, (float, int)))
+            other_ratio = other_count / len(observations)
+            raise ValueError(
+                f"Observer: {self.__class__.__name__} gathered observations with invalid values (not float or int)! "
+                f"Ratio: {other_ratio:.2%} ({other_count}/{len(observations)})"
+            )
+
+        # Check for NaN values
+        if any(np.isnan(observation) for observation in observations):
+            nan_count = sum(1 for observation in observations if np.isnan(observation))
+            nan_ratio = nan_count / len(observations)
+            logger.warning(
+                f"Observer: {self.__class__.__name__} gathered observations with NaN values! Ratio: {nan_ratio:.2%} "
+                f" ({nan_count}/{len(observations)})"
+            )
+            validate = False
+
+        # Check for very large values (including negative values)
+        if any(abs(observation) > self.invalid_observation_threshold for observation in observations):
+            inf_count = sum(1 for observation in observations if observation > self.invalid_observation_threshold)
+            inf_ratio = inf_count / len(observations)
+            logger.warning(
+                f"Observer: {self.__class__.__name__} gathered observations with large values (threshold: "
+                f"{self.invalid_observation_threshold})! Ratio: {inf_ratio:.2%} ({inf_count}/{len(observations)})"
+            )
+            validate = False
+
+        logger.trace(
+            f'Observer: {self.__class__.__name__} observation validation {"passed" if validate else "failed"}!'
+        )
+
     @abstractmethod
     def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
         """
@@ -204,16 +282,22 @@ class Observer(ABC):
         hyperparameter_states: HyperparameterStates,
         tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
         action_taken: float | int | None,
-    ) -> list[float | int]:
+        return_dict: bool = False,
+    ) -> tuple[list[float | int], dict[str, float] | None]:
         """
         :param observation_inputs: Observation input metrics not calculated with statistic trackers.
         :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
         :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
         names to floats or TensorStatistic models.
         :param action_taken: Action taken by the agent this class instance is assigned to.
-        :return: List of floats or integers to add to the agent's observation vector.
+        :param return_dict: Whether to return a dictionary of observations as well as the normal vector.
+        :return: Tuple of:
+            - List of floats or integers to add to the agent's observation vector.
+            - Dictionary of specific observation values if the storage type is TensorStatistics and None otherwise.
         """
 
+        observations_dict: dict[str, float] | None = None
+
         observations = self._observe(
             observation_inputs=observation_inputs,
             hyperparameter_states=hyperparameter_states,
@@ -225,7 +309,13 @@ class Observer(ABC):
             self._cached_observation = deepcopy(observations)
 
         if self.observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
-            observations = observations.to_list(skip_statistics=self.skip_statistics)  # type: ignore
+            if self.include_statistics is None:
+                raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")
+
+            if return_dict:
+                observations_dict = observations.as_observation_dict()  # type: ignore
+
+            observations = observations.to_list(include_statistics=self.include_statistics)  # type: ignore
 
         observations = [observations] if not isinstance(observations, list) else observations  # type: ignore
 
@@ -241,13 +331,15 @@ class Observer(ABC):
         if not self._validated_observation:
             self._validate_observation(observations=observations)
 
-        return observations
+        self._validate_observation_values(observations=observations)
+
+        return observations, observations_dict
 
     @final
     def inform(self) -> float | int | dict[str, float] | None:
         """
         :return: The cached observation. If the observation format is TensorStatistics then it is converted to a
-        dictionary with the statistics specified in skip_statistics excluded.
+        dictionary with the statistics specified in include_statistics included.
         """
 
         if not self.can_inform:
@@ -260,7 +352,10 @@ class Observer(ABC):
             )
 
         if self.observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
-            observation = self._cached_observation.model_dump(exclude=set(self.skip_statistics))  # type: ignore
+            if self.include_statistics is None:
+                raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")
+
+            observation = self._cached_observation.model_dump(include=set(self.include_statistics))  # type: ignore
 
         else:
             observation = self._cached_observation
