@@ -10,8 +10,8 @@ from typing import Any
 from libinephany.observations.observation_utils import StatisticStorageTypes
 from libinephany.observations.observers.base_observers import GlobalObserver
 from libinephany.observations.observers.global_observers.base_classes import (
+    LHOPTBaseObserver,
     LHOPTCheckpointBaseObserver,
-    LHOPTOuterStepBaseObserver,
 )
 from libinephany.observations.observers.global_observers.constants import LHOPT_CONSTANTS
 from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
@@ -229,7 +229,7 @@ class ValidationScore(GlobalObserver):
         return {}
 
 
-class LHOPTTrainingLoss(LHOPTOuterStepBaseObserver):
+class LHOPTTrainingLoss(LHOPTBaseObserver):
     """
     This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
     https://arxiv.org/abs/2305.18291.
@@ -239,6 +239,13 @@ class LHOPTTrainingLoss(LHOPTOuterStepBaseObserver):
     This observer use the CDF calculation from the paper and applies CDF transformation using the CDF mean and std.
     """
 
+    @property
+    def vector_length(self) -> int:
+        """
+        :return: Length of the vector returned by this observation if it returns a vector.
+        """
+        return 3  # [is_nan, is_inf, cdf_feature]
+
     def _observe(
         self,
         observation_inputs: ObservationInputs,
@@ -271,20 +278,8 @@ class LHOPTTrainingLoss(LHOPTOuterStepBaseObserver):
 
         return {}
 
-    def reset(self) -> None:
-        """Reset the observer by clearing the time series."""
-        self._time_series: list[tuple[float, float]] = []
-        self._current_time: float = 0.0
 
-    @property
-    def vector_length(self) -> int:
-        """
-        :return: Length of the vector returned by this observation if it returns a vector.
-        """
-        return 3  # [is_nan, is_inf, cdf_feature]
-
-
-class LHOPTValidationLoss(LHOPTOuterStepBaseObserver):
+class LHOPTValidationLoss(LHOPTBaseObserver):
     """
     This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
     https://arxiv.org/abs/2305.18291.
@@ -294,6 +289,13 @@ class LHOPTValidationLoss(LHOPTOuterStepBaseObserver):
     This observer use the CDF calculation from the paper and applies CDF transformation using the CDF mean and std.
     """
 
+    @property
+    def vector_length(self) -> int:
+        """
+        :return: Length of the vector returned by this observation if it returns a vector.
+        """
+        return 3  # [is_nan, is_inf, cdf_feature]
+
     def _observe(
         self,
         observation_inputs: ObservationInputs,
@@ -326,20 +328,8 @@ class LHOPTValidationLoss(LHOPTOuterStepBaseObserver):
 
         return {}
 
-    def reset(self) -> None:
-        """Reset the observer by clearing the time series."""
-        self._time_series: list[tuple[float, float]] = []
-        self._current_time: float = 0.0
 
-    @property
-    def vector_length(self) -> int:
-        """
-        :return: Length of the vector returned by this observation if it returns a vector.
-        """
-        return 3  # [is_nan, is_inf, cdf_feature]
-
-
-class LHOPTLossRatio(LHOPTOuterStepBaseObserver):
+class LHOPTLossRatio(LHOPTBaseObserver):
     """
     This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
     https://arxiv.org/abs/2305.18291.
@@ -353,6 +343,13 @@ class LHOPTLossRatio(LHOPTOuterStepBaseObserver):
     3. cdf_feature - CDF transformed feature using CDF mean and std
     """
 
+    @property
+    def vector_length(self) -> int:
+        """
+        :return: Length of the vector returned by this observation if it returns a vector.
+        """
+        return 3  # [is_nan, tanh, cdf_feature]
+
     def _observe(
         self,
         observation_inputs: ObservationInputs,
@@ -370,7 +367,7 @@ class LHOPTLossRatio(LHOPTOuterStepBaseObserver):
         """
 
         log_ratio = self._compute_log_ratio(
-            training_score=observation_inputs.training_score, validation_score=observation_inputs.validation_score
+            numerator=observation_inputs.training_loss, denominator=observation_inputs.validation_loss
         )
 
         tanh_feature = math.tanh(max(-LHOPT_CONSTANTS["TANH_BOUND"], min(LHOPT_CONSTANTS["TANH_BOUND"], log_ratio)))
@@ -381,31 +378,13 @@ class LHOPTLossRatio(LHOPTOuterStepBaseObserver):
 
         return [int(math.isnan(log_ratio)), tanh_feature, cdf_feature]
 
-    def _compute_log_ratio(self, training_score: float, validation_score: float) -> float:
+    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
         """
-        Compute the log ratio of validation_score to training_score.
-
-        :param training_score: Training score value
-        :param validation_score: Validation score value
-        :return: Log ratio value
+        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
+        needed.
         """
-        if training_score <= 0:
-            return 0.0
 
-        if validation_score <= 0:
-            return 0.0
-
-        # Calculate the ratio of validation_score to training_score
-        score_ratio = validation_score / training_score
-
-        return math.log(score_ratio)
-
-    @property
-    def vector_length(self) -> int:
-        """
-        :return: Length of the vector returned by this observation if it returns a vector.
-        """
-        return 3  # [is_nan, tanh, cdf_feature]
+        return {}
 
 
 class PercentileOfLossAtEachCheckpoint(LHOPTCheckpointBaseObserver):
