@@ -4,11 +4,13 @@
 #
 # ======================================================================================================================
 
+import math
 from typing import Any
 
 from libinephany.observations import observation_utils, statistic_trackers
-from libinephany.observations.observation_utils import StatisticStorageTypes
+from libinephany.observations.observation_utils import StatisticStorageTypes, compute_cdf_feature
 from libinephany.observations.observers.base_observers import LocalObserver
+from libinephany.observations.observers.global_observers.constants import LHOPT_CONSTANTS
 from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
 from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
 from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates
@@ -25,6 +27,14 @@ from libinephany.utils.transforms import HyperparameterTransformType
 
 class FirstOrderGradients(LocalObserver):
 
+    @property
+    def requires_include_statistics(self) -> bool:
+        """
+        :return: Whether the observation requires include_statistics to be provided.
+        """
+
+        return True
+
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
         :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
@@ -66,7 +76,7 @@ class FirstOrderGradients(LocalObserver):
         needed.
         """
 
-        return {statistic_trackers.FirstOrderGradients.__name__: dict(skip_statistics=self.skip_statistics)}
+        return {statistic_trackers.FirstOrderGradients.__name__: dict(include_statistics=self.include_statistics)}
 
 
 class SecondOrderGradients(LocalObserver):
@@ -87,6 +97,14 @@ class SecondOrderGradients(LocalObserver):
 
         self.compute_hessian_diagonal = compute_hessian_diagonal
 
+    @property
+    def requires_include_statistics(self) -> bool:
+        """
+        :return: Whether the observation requires include_statistics to be provided.
+        """
+
+        return True
+
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
         :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
@@ -130,13 +148,21 @@ class SecondOrderGradients(LocalObserver):
 
         return {
             statistic_trackers.SecondOrderGradients.__name__: dict(
-                skip_statistics=self.skip_statistics, compute_hessian_diagonal=self.compute_hessian_diagonal
+                include_statistics=self.include_statistics, compute_hessian_diagonal=self.compute_hessian_diagonal
             )
         }
 
 
 class Activations(LocalObserver):
 
+    @property
+    def requires_include_statistics(self) -> bool:
+        """
+        :return: Whether the observation requires include_statistics to be provided.
+        """
+
+        return True
+
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
         :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
@@ -178,11 +204,19 @@ class Activations(LocalObserver):
         needed.
         """
 
-        return {statistic_trackers.ActivationStatistics.__name__: dict(skip_statistics=self.skip_statistics)}
+        return {statistic_trackers.ActivationStatistics.__name__: dict(include_statistics=self.include_statistics)}
 
 
 class ParameterUpdates(LocalObserver):
 
+    @property
+    def requires_include_statistics(self) -> bool:
+        """
+        :return: Whether the observation requires include_statistics to be provided.
+        """
+
+        return True
+
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
         :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
@@ -224,11 +258,19 @@ class ParameterUpdates(LocalObserver):
         needed.
         """
 
-        return {statistic_trackers.ParameterUpdateStatistics.__name__: dict(skip_statistics=self.skip_statistics)}
+        return {statistic_trackers.ParameterUpdateStatistics.__name__: dict(include_statistics=self.include_statistics)}
 
 
 class Parameters(LocalObserver):
 
+    @property
+    def requires_include_statistics(self) -> bool:
+        """
+        :return: Whether the observation requires include_statistics to be provided.
+        """
+
+        return True
+
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
         :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
@@ -270,7 +312,7 @@ class Parameters(LocalObserver):
         needed.
         """
 
-        return {statistic_trackers.ParameterStatistics.__name__: dict(skip_statistics=self.skip_statistics)}
+        return {statistic_trackers.ParameterStatistics.__name__: dict(include_statistics=self.include_statistics)}
 
 
 class LAMBTrustRatio(LocalObserver):
@@ -469,6 +511,94 @@ class ActionSchemeOneHot(LocalObserver):
         return {}
 
 
+class PreviousActionRepresentation(LocalObserver):
+    """
+    Observer that returns the representation of the previous action taken by the agent.
+
+    This observer tracks the previous action and returns it in an appropriate format:
+    - For discrete actions: returns one-hot encoding of the previous action
+    - For continuous actions: returns the previous action value directly
+    """
+
+    DISCRETE_INDEX = 0
+
+    def __init__(self, **kwargs) -> None:
+        super().__init__(**kwargs)
+        self._previous_action: float | int | None = None
+
+    @property
+    def vector_length(self) -> int:
+        """
+        :return: Length of the vector returned by this observation if it returns a vector.
+        """
+        return self.number_of_discrete_actions if self.is_discrete else 1
+
+    @property
+    def is_discrete(self) -> bool:
+        """
+        :return: Whether the agent is using discrete actions.
+        """
+
+        valid_actions = self.number_of_discrete_actions is not None and self.number_of_discrete_actions > 0
+        return self.action_scheme_index == self.DISCRETE_INDEX and valid_actions
+
+    @property
+    def can_inform(self) -> bool:
+        """
+        :return: Whether observations from the observer can be used in the agent info dictionary.
+        """
+        return False
+
+    def _get_observation_format(self) -> StatisticStorageTypes:
+        """
+        :return: Format the observation returns data in.
+        """
+        return StatisticStorageTypes.VECTOR
+
+    def _observe(
+        self,
+        observation_inputs: ObservationInputs,
+        hyperparameter_states: HyperparameterStates,
+        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
+        action_taken: float | int | None,
+    ) -> float | int | list[int | float] | TensorStatistics:
+        """
+        Returns the representation of the previous action.
+
+        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
+        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
+        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
+        names to floats or TensorStatistic models.
+        :param action_taken: Action taken by the agent this class instance is assigned to.
+        :return: Previous action representation (one-hot vector for discrete, float for continuous).
+        """
+
+        if self._previous_action is None:
+            result = [0.0] * self.vector_length
+        else:
+            if self.is_discrete:
+                result = observation_utils.create_one_hot_observation(
+                    vector_length=self.vector_length, one_hot_index=int(self._previous_action)
+                )
+            else:
+                result = [float(self._previous_action)]
+
+        self._previous_action = action_taken
+
+        return result
+
+    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
+        """
+        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
+        needed.
+        """
+        return {}
+
+    def reset(self) -> None:
+        """Resets the observer by clearing the previous action."""
+        self._previous_action = None
+
+
 class DepthOneHot(LocalObserver):
 
     @property
@@ -534,6 +664,7 @@ class ModuleTypeOneHot(LocalObserver):
         "attention": 1,
         "linear": 2,
         "embedding": 3,
+        "lstm": 4,
     }
 
     @property
@@ -552,6 +683,14 @@ class ModuleTypeOneHot(LocalObserver):
 
         return False
 
+    @property
+    def can_standardize(self) -> bool:
+        """
+        :return: Whether the observation can be standardized.
+        """
+
+        return False
+
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
         :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
@@ -602,17 +741,6 @@ class ModuleTypeOneHot(LocalObserver):
 
 class CurrentHyperparameters(LocalObserver):
 
-    def __init__(self, skip_hparams: list[str] | None = None, **kwargs) -> None:
-        """
-        :param skip_hparams: Names of the hyperparameters to not include in the initial values vector returned by
-        this observation.
-        :param kwargs: Miscellaneous keyword arguments.
-        """
-
-        super().__init__(**kwargs)
-
-        self.skip_hparams = skip_hparams if skip_hparams is not None else []
-
     @property
     def can_standardize(self) -> bool:
         """
@@ -635,11 +763,20 @@ class CurrentHyperparameters(LocalObserver):
         :return: Length of the vector returned by this observation if it returns a vector.
         """
 
+        if self.include_hparams is None:
+            raise ValueError(f"{self.__class__.__name__} must be provided with include_hparams.")
+
         available_hparams = HyperparameterStates.get_layerwise_hyperparameters()
 
-        return len(
-            [hparam for hparam in available_hparams if not any(skipped in hparam for skipped in self.skip_hparams)]
-        )
+        return len([hparam for hparam in available_hparams if hparam in self.include_hparams])
+
+    @property
+    def requires_include_hparams(self) -> bool:
+        """
+        :return: Whether the observation requires include_hparams to be provided.
+        """
+
+        return True
 
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
@@ -668,7 +805,7 @@ class CurrentHyperparameters(LocalObserver):
         assert self.parameter_group_name is not None
 
         current_internal_values = hyperparameter_states[self.parameter_group_name].get_current_internal_values(
-            skip_hparams=self.skip_hparams
+            include_hparams=self.include_hparams
         )
 
         self._cached_observation = current_internal_values
@@ -686,17 +823,6 @@ class CurrentHyperparameters(LocalObserver):
 
 class CurrentHyperparameterDeltas(LocalObserver):
 
-    def __init__(self, skip_hparams: list[str] | None = None, **kwargs) -> None:
-        """
-        :param skip_hparams: Names of the hyperparameters to not include in the initial deltas vector returned by
-        this observation.
-        :param kwargs: Miscellaneous keyword arguments.
-        """
-
-        super().__init__(**kwargs)
-
-        self.skip_hparams = skip_hparams if skip_hparams is not None else []
-
     @property
     def can_standardize(self) -> bool:
         """
@@ -719,11 +845,20 @@ class CurrentHyperparameterDeltas(LocalObserver):
         :return: Length of the vector returned by this observation if it returns a vector.
         """
 
+        if self.include_hparams is None:
+            raise ValueError(f"{self.__class__.__name__} must be provided with include_hparams.")
+
         available_hparams = HyperparameterStates.get_layerwise_hyperparameters()
 
-        return len(
-            [hparam for hparam in available_hparams if not any(skipped in hparam for skipped in self.skip_hparams)]
-        )
+        return len([hparam for hparam in available_hparams if hparam in self.include_hparams])
+
+    @property
+    def requires_include_hparams(self) -> bool:
+        """
+        :return: Whether the observation requires include_hparams to be provided.
+        """
+
+        return True
 
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
@@ -750,9 +885,10 @@ class CurrentHyperparameterDeltas(LocalObserver):
         """
 
         assert self.parameter_group_name is not None
+        assert self.include_hparams is not None
 
         current_deltas = hyperparameter_states[self.parameter_group_name].get_current_deltas(
-            skip_hparams=self.skip_hparams
+            include_hparams=self.include_hparams
         )
 
         self._cached_observation = current_deltas
@@ -772,17 +908,6 @@ class HyperparameterTransformTypes(LocalObserver):
 
     TRANSFORM_TYPE_TO_IDX = dict(((s, i) for i, s in enumerate(HyperparameterTransformType)))
 
-    def __init__(self, skip_hparams: list[str] | None = None, **kwargs) -> None:
-        """
-        :param skip_hparams: Names of the hyperparameters to not include in the transforms vector returned by
-        this observation.
-        :param kwargs: Miscellaneous keyword arguments.
-        """
-
-        super().__init__(**kwargs)
-
-        self.skip_hparams = skip_hparams if skip_hparams is not None else []
-
     @property
     def can_standardize(self) -> bool:
         """
@@ -805,12 +930,23 @@ class HyperparameterTransformTypes(LocalObserver):
         :return: Length of the vector returned by this observation if it returns a vector.
         """
 
+        if self.include_hparams is None:
+            raise ValueError(f"{self.__class__.__name__} must be provided with include_hparams.")
+
         available_hparams = HyperparameterStates.get_layerwise_hyperparameters()
 
         return len(HyperparameterTransformType) * len(
-            [hparam for hparam in available_hparams if not any(skipped in hparam for skipped in self.skip_hparams)]
+            [hparam for hparam in available_hparams if hparam in self.include_hparams]
         )
 
+    @property
+    def requires_include_hparams(self) -> bool:
+        """
+        :return: Whether the observation requires include_hparams to be provided.
+        """
+
+        return True
+
     def _get_observation_format(self) -> StatisticStorageTypes:
         """
         :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
@@ -836,10 +972,11 @@ class HyperparameterTransformTypes(LocalObserver):
         """
 
         assert self.parameter_group_name is not None
+        assert self.include_hparams is not None
 
         parameter_group_hparams = hyperparameter_states[self.parameter_group_name]
         hyperparameter_transform_types = parameter_group_hparams.get_hyperparameter_transform_types(
-            skip_hparams=self.skip_hparams
+            include_hparams=self.include_hparams
         )
         hyperparameter_transform_types_onehot_list = [
             observation_utils.create_one_hot_observation(
@@ -981,3 +1118,133 @@ class PercentageDepth(LocalObserver):
         """
 
         return {}
+
+
+class LogOfNoiseScaleObserver(LocalObserver):
+
+    def __init__(
+        self,
+        *,
+        decay_factor: float = LHOPT_CONSTANTS["DEFAULT_DECAY_FACTOR"],
+        time_window: int = LHOPT_CONSTANTS["DEFAULT_TIME_WINDOW"],
+        **kwargs,
+    ) -> None:
+        """
+        :param decay_factor: Decay factor for CDF calculation in [1, 2.5, 5, 10, 20]
+        :param time_window: Number of time steps to consider for CDF calculation
+        :param include_statistics: List of statistics to include.
+        or use the squared first order gradients as approximations in the same way Adam does.
+        :param kwargs: Miscellaneous keyword arguments.
+        """
+
+        super().__init__(**kwargs)
+
+        self.decay_factor = max(0.0, decay_factor)
+        self.time_window = max(1, time_window)
+
+        # Store time series data for CDF calculation
+        self._time_series: list[tuple[float, float]] = []  # (time, value) pairs
+        self._current_time: float = 0.0
+
+    @property
+    def can_standardize(self) -> bool:
+        """
+        :return: Whether the observation can be standardized.
+        """
+
+        return False
+
+    @property
+    def can_inform(self) -> bool:
+        """
+        :return: Whether observations from the observer can be used in the agent info dictionary.
+        """
+
+        return False
+
+    @property
+    def vector_length(self) -> int:
+        """
+        :return: Length of the vector returned by this observation if it returns a vector.
+        """
+        return 2  # [log_noise_scale, cdf_feature]
+
+    @property
+    def requires_include_statistics(self) -> bool:
+        """
+        :return: Whether the observation requires include_statistics to be provided.
+        """
+
+        return True
+
+    def _get_observation_format(self) -> StatisticStorageTypes:
+        """
+        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
+        enumeration class.
+        """
+
+        return StatisticStorageTypes.VECTOR
+
+    def _update_time(self) -> None:
+        """Update the current time counter."""
+        self._current_time += 1.0
+
+    def _compute_cdf_feature(self, value: float) -> float:
+        """
+        Compute CDF feature for the given value.
+        training loss will be added to the time series after this call.
+        :param value: The value to compute CDF feature for
+        :return: CDF feature value
+        """
+        return compute_cdf_feature(value, self._time_series, self.decay_factor, self._current_time, self.time_window)
+
+    def _observe(
+        self,
+        observation_inputs: ObservationInputs,
+        hyperparameter_states: HyperparameterStates,
+        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
+        action_taken: float | int | None,
+    ) -> float | int | list[int | float] | TensorStatistics:
+        """
+        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
+        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
+        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
+        names to floats or TensorStatistic models.
+        :param action_taken: Action taken by the agent this class instance is assigned to.
+        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
+        """
+
+        statistics = tracked_statistics[statistic_trackers.LogOfNoiseScaleStatistics.__name__]
+        raw_value = list(statistics.values())[0]  # type: ignore[list-item]
+
+        assert isinstance(raw_value, float), f"Expected float, got {type(raw_value)}"  # to avoid type errors with mypy
+
+        batch_size = hyperparameter_states.global_hparams.batch_size.external_value
+        learning_rate = hyperparameter_states.parameter_group_hparams[
+            self.parameter_group_name
+        ].learning_rate.external_value
+
+        log_b_over_epsilon = math.log(batch_size / learning_rate)
+        log_noise_scale = raw_value + log_b_over_epsilon
+
+        cdf_feature = self._compute_cdf_feature(log_noise_scale)  # type: ignore[arg-type]
+        self._update_time()
+
+        return [log_noise_scale, cdf_feature]  # type: ignore[list-item]
+
+    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
+        """
+        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
+        needed.
+        """
+
+        return {
+            statistic_trackers.LogOfNoiseScaleStatistics.__name__: dict(
+                include_statistics=self.include_statistics, sample_frequency=LHOPT_CONSTANTS["DEFAULT_SAMPLE_FREQUENCY"]
+            )
+        }
+
+    def reset(self) -> None:
+        """Reset the observer by clearing the time series."""
+        self._time_series = []
+        self._current_time = 0.0
