@@ -134,7 +134,7 @@ class ObserverPipelineCoordinator:
 
         clipped_observations = False
         observations = {}
-        observations_as_dict: dict[str, dict[str, list[float | int]]] = {}
+        observations_as_dict: dict[str, dict[str, list[float | int] | dict[str, float | int]]] = {}
 
         for pipeline in self.pipelines:
             pipeline_observations, pipeline_clipped_observations, pipeline_observations_dict = pipeline.observe(
