@@ -51,7 +51,7 @@ class Observations(BaseModel):
     hit_invalid_value: bool
 
     agent_observations: dict[str, list[float | int]]
-    observations_as_dict: dict[str, dict[str, list[float | int]]] | None = None
+    observations_as_dict: dict[str, dict[str, list[float | int] | dict[str, float]]] | None = None
 
     def observations_as_arrays(self, dtype: DTypeLike = np.float32) -> dict[str, np.ndarray]:
         """
