@@ -60,9 +60,7 @@ class UpdateCallbacks(BaseModel):
     adam_eps: Callable[..., None]
     sgd_momentum: Callable[..., None]
 
-    batch_size: Callable[..., None] | None
     gradient_accumulation: Callable[..., None] | None
-    epochs: Callable[..., None] | None
 
     def __getitem__(self, item: str) -> Callable[..., None] | None:
         """
@@ -295,52 +293,52 @@ class HyperparameterContainer(BaseModel):
             f"{self.get_hyperparameters()}."
         )
 
-    def get_initial_internal_values(self, skip_hparams: list[str] | None = None) -> dict[str, float | int | None]:
+    def get_initial_internal_values(self, include_hparams: list[str] | None = None) -> dict[str, float | int | None]:
         """
-        :param skip_hparams: Names of the hyperparameters to not include in the initial values vector returned by
-        this observation.
+        :param include_hparams: Names of the hyperparameters to include in the hyperparameter name to initial internal
+        value mapping.
         :return: Dictionary mapping hyperparameter names to their current values for this parameter group.
         """
 
-        if skip_hparams is None:
-            skip_hparams = []
+        if include_hparams is None:
+            include_hparams = list(self.model_fields.keys())
 
         return {
             field_name: field_value.initial_internal_value
             for field_name, field_value in self.__dict__.items()
-            if isinstance(field_value, Hyperparameter) and field_name not in skip_hparams
+            if isinstance(field_value, Hyperparameter) and field_name in include_hparams
         }
 
-    def get_current_internal_values(self, skip_hparams: list[str] | None = None) -> dict[str, float | int | None]:
+    def get_current_internal_values(self, include_hparams: list[str] | None = None) -> dict[str, float | int | None]:
         """
-        :param skip_hparams: Names of the hyperparameters to not include in the current values vector returned by
-        this observation.
+        :param include_hparams: Names of the hyperparameters to include in the hyperparameter name to current internal
+        value mapping.
         :return: Dictionary mapping hyperparameter names to their current values for this parameter group.
         """
 
-        if skip_hparams is None:
-            skip_hparams = []
+        if include_hparams is None:
+            include_hparams = list(self.model_fields.keys())
 
         return {
             field_name: field_value.current_internal_value
             for field_name, field_value in self.__dict__.items()
-            if isinstance(field_value, Hyperparameter) and field_name not in skip_hparams
+            if isinstance(field_value, Hyperparameter) and field_name in include_hparams
         }
 
-    def get_current_deltas(self, skip_hparams: list[str] | None = None) -> dict[str, float | int | None]:
+    def get_current_deltas(self, include_hparams: list[str] | None = None) -> dict[str, float | int | None]:
         """
-        :param skip_hparams: Names of the hyperparameters to not include in the current deltas vector returned by
-        this observation.
+        :param include_hparams: Names of the hyperparameters to include in the hyperparameter name to current delta
+        mapping.
         :return: Dictionary mapping hyperparameter names to their current deltas for this parameter group.
         """
 
-        if skip_hparams is None:
-            skip_hparams = []
+        if include_hparams is None:
+            include_hparams = list(self.model_fields.keys())
 
         return {
             field_name: field_value.current_delta
             for field_name, field_value in self.__dict__.items()
-            if isinstance(field_value, Hyperparameter) and field_name not in skip_hparams
+            if isinstance(field_value, Hyperparameter) and field_name in include_hparams
         }
 
     def set_internal_values(self, internal_values: dict[str, float | int | None]) -> None:
@@ -440,19 +438,21 @@ class ParameterGroupHParams(HyperparameterContainer):
         )
 
     def get_hyperparameter_transform_types(
-        self, skip_hparams: list[str] | None = None
+        self, include_hparams: list[str] | None = None
     ) -> dict[str, HyperparameterTransformType]:
         """
+        :param include_hparams: Names of the hyperparameters to include in the hyperparameter name to transform type
+        mapping.
         :return: Dictionary mapping hyperparameter names to their transform type for this parameter group.
         """
 
-        if skip_hparams is None:
-            skip_hparams = []
+        if include_hparams is None:
+            include_hparams = list(self.model_fields.keys())
 
         return {
             field_name: field_value.transform_type
             for field_name, field_value in self.__dict__.items()
-            if isinstance(field_value, Hyperparameter) and field_name not in skip_hparams
+            if isinstance(field_value, Hyperparameter) and field_name in include_hparams
         }
 
 
@@ -839,15 +839,15 @@ class HyperparameterStates(BaseModel):
         for parameter_group_hparams in self.parameter_group_hparams.values():
             parameter_group_hparams.set_to_initial_values()
 
-    def get_initial_internal_values(self, skip_hparams: list[str] | None = None) -> dict[str, float | int | None]:
+    def get_initial_internal_values(self, include_hparams: list[str] | None = None) -> dict[str, float | int | None]:
         """
-        :param skip_hparams: Hyperparameters to ignore while retrieving initial values.
+        :param include_hparams: Hyperparameters to include while retrieving initial values.
         :return: Dictionary mapping hyperparameter names to their initial values at the start of training.
         """
 
         initial_internal_values = {
-            **self.global_hparams.get_initial_internal_values(skip_hparams),
-            **next(iter(self.parameter_group_hparams.values())).get_initial_internal_values(skip_hparams),
+            **self.global_hparams.get_initial_internal_values(include_hparams),
+            **next(iter(self.parameter_group_hparams.values())).get_initial_internal_values(include_hparams),
         }
         initial_internal_values = {
             hparam_name: initial_internal_values.get(hparam_name, None)
@@ -855,3 +855,20 @@ class HyperparameterStates(BaseModel):
         }
 
         return initial_internal_values
+
+    def get_current_internal_values(self, include_hparams: list[str] | None = None) -> dict[str, float | int | None]:
+        """
+        :param include_hparams: Hyperparameters to include while retrieving current values.
+        :return: Dictionary mapping hyperparameter names to their current values during training.
+        """
+
+        current_internal_values = {
+            **self.global_hparams.get_current_internal_values(include_hparams),
+            **next(iter(self.parameter_group_hparams.values())).get_current_internal_values(include_hparams),
+        }
+        current_internal_values = {
+            hparam_name: current_internal_values.get(hparam_name, None)
+            for hparam_name in self.initial_hyperparameter_internal_values
+        }
+
+        return current_internal_values
