@@ -8,6 +8,7 @@ KEY_HEADER_CASE = "X-API-Key"
 KEY_HEADER_NO_CASE = KEY_HEADER_CASE.lower()
 
 TIMESTAMP_FORMAT = "%Y-%m-%d-%H-%M-%S"
+TIMESTAMP_FORMAT_WITH_MS = "%Y-%m-%d-%H-%M-%S-%f"
 
 RLLIB_TRUNC_EPISODES = "truncate_episodes"
 RLLIB_COMP_EPISODES = "complete_episodes"
@@ -42,7 +43,7 @@ AGENT_PREFIX_EPS = "adam-eps"
 AGENT_PREFIX_SGD_MOMENTUM = "sgd-momentum"
 
 AGENT_BATCH_SIZE = "batch-size"
-AGENT_GRADIENT_ACCUMULATION = "gradient-accumulation"
+AGENT_PREFIX_GRADIENT_ACCUMULATION = "gradient-accumulation"
 
 AGENT_BANDIT_SUFFIX = "bandit-agent"
 
@@ -67,7 +68,7 @@ PREFIXES = [
     AGENT_PREFIX_BETA_TWO,
     AGENT_PREFIX_EPS,
     AGENT_PREFIX_SGD_MOMENTUM,
-    AGENT_GRADIENT_ACCUMULATION,
+    AGENT_PREFIX_GRADIENT_ACCUMULATION,
 ]
 PREFIXES_TO_HPARAMS = {
     AGENT_PREFIX_LR: LEARNING_RATE,
@@ -78,6 +79,6 @@ PREFIXES_TO_HPARAMS = {
     AGENT_PREFIX_BETA_TWO: ADAM_BETA_TWO,
     AGENT_PREFIX_EPS: ADAM_EPS,
     AGENT_PREFIX_SGD_MOMENTUM: SGD_MOMENTUM,
-    AGENT_GRADIENT_ACCUMULATION: GRADIENT_ACCUMULATION,
+    AGENT_PREFIX_GRADIENT_ACCUMULATION: GRADIENT_ACCUMULATION,
 }
 HPARAMS_TO_PREFIXES = {hparam: prefix for prefix, hparam in PREFIXES_TO_HPARAMS.items()}
