@@ -58,6 +58,8 @@ class ErrorLogger(AsyncioWorker):
     def __init__(
         self,
         service_name: str,
+        member_ids_to_mention: list[str] | None = None,
+        custom_error_message: str | None = None,
         minimum_severity: ErrorSeverities = ErrorSeverities.WARNING,
         error_classifier: ExceptionSeverityClassifier | None = None,
         max_time_accumulating_errors: float = 10.0,
@@ -66,6 +68,8 @@ class ErrorLogger(AsyncioWorker):
         """
         :param service_name: Name of the service deployed on the cloud errors sent to this error logger should be sent
         to.
+        :param member_ids_to_mention: List of member IDs to mention in the Slack message. If None, no members will be mentioned.
+        :param custom_error_message: Custom error message to include in the Slack message. If None, no custom message will be included.
         :param minimum_severity: Minimum error severity that can be sent to Slack.
         :param error_classifier: Optional callable used to alter how errors severities are classified. If this is None
         all errors are given the 'WARNING' severity.
@@ -77,6 +81,8 @@ class ErrorLogger(AsyncioWorker):
 
         self._slack_client, self._channel_id = self._get_slack_client()
 
+        self.member_mentions = self._form_member_mentions(member_ids_to_mention=member_ids_to_mention)
+        self.custom_error_message = custom_error_message
         self.service_name = service_name
         self.error_classifier = error_classifier if error_classifier is not None else default_error_classifier
         self.max_time_accumulating_errors = max_time_accumulating_errors
@@ -144,6 +150,7 @@ class ErrorLogger(AsyncioWorker):
         exception: Exception,
         frequency: int,
         frequency_time_window: float,
+        custom_error_message: str | None,
     ) -> str:
         """
         :param mention: Mention tag of the entire channel or a particular user ID.
@@ -152,19 +159,40 @@ class ErrorLogger(AsyncioWorker):
         :param exception: Exception that occurred.
         :param frequency: How frequently the exception occurred in the given time window.
         :param frequency_time_window: Time between the first and last instance of the exception.
+        :param custom_error_message: Custom error message to include in the Slack message. If None, no custom message will be included.
         :return: Formatted Slack message string.
         """
 
         frequency_line = f"{frequency} in {frequency_time_window}s" if frequency > 1 else f"{frequency}"
 
+        header = f"{notification_emoji} *{service_name} Error* {notification_emoji}\n"
+
+        if mention:
+            header += f"• *Alerting*: {mention}\n"
+
+        if custom_error_message is not None:
+            header += f"• *Custom Message*: {custom_error_message}\n"
+
         return (
-            f"{mention}{notification_emoji} *{service_name} Error* {notification_emoji}\n"
+            f"{header}"
             f"• *Service*: {service_name}\n"
             f"• *Error Type*: {exception.__class__.__name__}\n"
             f"• *Frequency*: {frequency_line}\n"
             f"• *Traceback*: "
         )
 
+    @staticmethod
+    def _form_member_mentions(member_ids_to_mention: list[str] | None) -> str:
+        """
+        :param member_ids_to_mention: List of member IDs to mention in the Slack message.
+        :return: String of member mentions.
+        """
+
+        if not member_ids_to_mention:
+            return ""
+
+        return " ".join([f"<@{member_id}>" for member_id in member_ids_to_mention])
+
     def _get_traceback_file_name(self, exception: Exception, exception_timestamp: str) -> str:
         """
         :param exception: Exception being sent to Slack.
@@ -275,7 +303,10 @@ class ErrorLogger(AsyncioWorker):
         traceback_filename = self._get_traceback_file_name(exception=exception, exception_timestamp=formatted_timestamp)
 
         notification_emoji = SEVERITY_EMOJIS[severity]
-        mention = "" if not ErrorSeverities.should_mention(severity=severity) else self.CHANNEL_MENTION
+        # Temporary since the bot is used in training and the API.
+        mention = (
+            self.member_mentions if not ErrorSeverities.should_mention(severity=severity) else self.CHANNEL_MENTION
+        )
 
         self._log_error_to_terminal(exception=exception)
 
@@ -291,6 +322,7 @@ class ErrorLogger(AsyncioWorker):
                     exception=exception,
                     frequency=frequency,
                     frequency_time_window=frequency_time_window,
+                    custom_error_message=self.custom_error_message,
                 ),
             )
         except SlackApiError as e:
